@@ -2,16 +2,26 @@
 
 import asyncio
 import logging
-from collections import defaultdict
+from contextlib import asynccontextmanager
 from datetime import datetime
 from functools import partial, lru_cache
 from itertools import chain
 from operator import itemgetter
+from time import time
 from typing import MutableMapping, List, Union, Coroutine, Type, Sequence, Dict, Callable, Awaitable, Optional, \
     Iterator, NoReturn, Tuple
 
+from patchwork.core.client.task import FrozenTask, Task
+
+try:
+    from prometheus_client import Counter, Gauge, Histogram
+except ImportError:
+    # prometheus not installed, use stubs instead
+    from patchwork.core.stubs.prometheus import Counter, Gauge, Histogram
+
+
 import pytz as pytz
-from patchwork.core import Task, AsyncSubscriber, TaskMetadata
+from patchwork.core import AsyncSubscriber, TaskMetadata
 from patchwork.core.config.base import ComponentConfig
 from patchwork.core.typing import FuncPath
 from patchwork.core.utils import cached_property, Flag
@@ -23,19 +33,30 @@ from patchwork.node.executor.unit.base import ProcessingUnit
 
 # task finalizer, run when task processing completes, if third argument is None
 # means that task has been not processed as there was no processor to handle it
-FinalizerType = Callable[[Task, AsyncSubscriber, asyncio.Future], Awaitable]
+FinalizerType = Callable[[FrozenTask, AsyncSubscriber, asyncio.Future], Awaitable]
 # middleware is to prepare task processing environment, optional finalizer is a method which
 # will be called to finalize task processing. Middleware may return new finalizer which should
 # invoke given finalizer internally for proper task finalization
 # if middleware does not return new finalizer, passed finalizer will be used
 # middlewares are run on the worker async loop, before passing task to processing unit!
 # keep this in mind as middleware execution may slow down whole executor
-MiddlewareType = Callable[[Task, AsyncSubscriber, FinalizerType], Awaitable[Optional[FinalizerType]]]
+MiddlewareType = Callable[[FrozenTask, AsyncSubscriber, FinalizerType], Awaitable[Optional[FinalizerType]]]
 
-ExceptionHandler = Callable[[Task, Exception], Awaitable]
+ExceptionHandler = Callable[[FrozenTask, Exception], Awaitable]
+
+# prometheus metrics
+tasks_processed = Counter('patchwork_node_tasks_total', "Number of tasks processed", ['task_type'])
+tasks_success = Counter('patchwork_node_success_total', "Number of tasks succeeded", ['task_type'])
+tasks_failed = Counter('patchwork_node_failed_total', "Number of tasks failed", ['task_type'])
+tasks_cancelled = Counter('patchwork_node_cancelled_total', "Number of tasks cancelled", ['task_type'])
+tasks_backoffed = Counter('patchwork_node_backoff_total', "Number of tasks backoffed", ['task_type'])
+tasks_in_progress = Gauge('patchwork_node_in_progress', "Number of tasks in progress", ['task_type'])
+tasks_time = Histogram('patchwork_node_tasks_time', "Tasks processing time", ['task_type'])
+tasks_wait_get_time = Histogram('patchwork_node_get_time', "Task fetching time")
+tasks_wait_unit_time = Histogram('patchwork_node_unit_time', "Task waiting time for processing unit slot")
 
 
-async def check_task_middleware(task: Task, receiver: AsyncSubscriber, finalizer: FinalizerType):
+async def check_task_middleware(task: FrozenTask, receiver: AsyncSubscriber, finalizer: FinalizerType):
     now = datetime.now(pytz.UTC)
     if task.meta.expires and task.meta.expires < now:
         # task has expired
@@ -45,7 +66,7 @@ async def check_task_middleware(task: Task, receiver: AsyncSubscriber, finalizer
         raise TaskRetry(countdown=(now - task.meta.not_before).total_seconds())
 
 
-async def default_task_exception_handler(task: Task, exception: Exception) -> NoReturn:
+async def default_task_exception_handler(task: FrozenTask, exception: Exception) -> NoReturn:
     if isinstance(exception, TaskControlException):
         raise exception
 
@@ -60,7 +81,7 @@ class Idempotent:
     def __init__(self, fn):
         self._fn = fn
 
-    def __call__(self, task: Task, exc: Exception):
+    def __call__(self, task: FrozenTask, exc: Exception):
         try:
             return self._fn(task, exc)
         except TaskControlException:
@@ -281,8 +302,9 @@ class BaseExecutor(Module):
         while True:
 
             try:
-                done, pending = await asyncio.wait((asyncio.create_task(subscriber.get()), term_future,),
-                                                   return_when=asyncio.FIRST_COMPLETED)
+                with tasks_wait_get_time.time():
+                    done, pending = await asyncio.wait((asyncio.create_task(subscriber.get()), term_future,),
+                                                       return_when=asyncio.FIRST_COMPLETED)
             except Exception as e:
                 self.logger.error(f"Client get() raises exception: {e.__class__.__name__}({e})",
                                   exc_info=True)
@@ -317,7 +339,7 @@ class BaseExecutor(Module):
                               f"{exc.__class__.__name__}({exc})")
             return True
 
-        task: Task = future.result()
+        task: FrozenTask = future.result()
         self.logger.debug(f"Client {receiver} waiter returned a task message: {task.uuid}")
 
         self._busy.set(True)
@@ -336,7 +358,7 @@ class BaseExecutor(Module):
 
         return True
 
-    async def retry(self, task: Task, *, countdown: float = None,
+    async def retry(self, task: FrozenTask, *, countdown: float = None,
                     not_before: datetime = None, bump_attempts: bool = True) -> Union[Task, None]:
         """
         Retries given task by cloning it and sending through publisher.
@@ -376,7 +398,7 @@ class BaseExecutor(Module):
         await publisher.send_task(new_task)
         return new_task
 
-    async def backoff_task(self, task, reason: str = None):
+    async def backoff_task(self, task: FrozenTask, reason: str = None):
         """
         Stores given task on backoff queue. If task cannot be executed for some reason (eg retries limit exceeded)
         can be logged and saved in special place for future investigation. This method allows to notify administrator
@@ -388,13 +410,14 @@ class BaseExecutor(Module):
         :param reason:
         :return:
         """
+        tasks_backoffed.labels(task_type=task.task_type).inc()
         backoff_log = logging.getLogger('backoff')
         backoff_log.log(logging.NOTSET,
                         f"Task '{task.uuid}' backoffed.{(' Reason: ' + reason) if reason is not None else ''}",
                         extra={'task': task})
         self.logger.debug(f"Task '{task.uuid}' backoffed.{(' Reason: ' + reason) if reason is not None else ''}")
 
-    async def handle(self, task: Task, receiver: AsyncSubscriber, task_future: asyncio.Future):
+    async def handle(self, task: FrozenTask, receiver: AsyncSubscriber, task_future: asyncio.Future):
         """
         Executes given task, handles all exceptions which may rise during task
         processing.
@@ -412,6 +435,8 @@ class BaseExecutor(Module):
         :param task_future: a future associated with the task which should resolved when task processing completes
         """
         finalizer = self._handle_task_result
+        tasks_in_progress.labels(task_type=task.task_type).inc()
+        start_time = time()
 
         try:
             for middleware in self._middlewares:
@@ -425,7 +450,13 @@ class BaseExecutor(Module):
         finalizer = partial(finalizer, task=task, receiver=receiver)
 
         exec_fut.add_done_callback(
-            partial(self._task_done_callback, finalizer=finalizer(processing_fut=exec_fut), task_future=task_future)
+            partial(
+                self._task_done_callback,
+                finalizer=finalizer(processing_fut=exec_fut),
+                task_future=task_future,
+                task=task,
+                start_time=start_time
+            )
         )
 
     def _task_handling_completed(self, task_future: asyncio.Future):
@@ -453,7 +484,8 @@ class BaseExecutor(Module):
             processor = router.handle(task, receiver)
             if processor is not None:
                 # await for submit which may hold if processing unit is full
-                pu_fut = await self.unit.submit(processor)
+                with tasks_wait_unit_time.time():
+                    pu_fut = await self.unit.submit(processor)
                 return pu_fut
 
         self.logger.error(f"No route to handle message type '{task.task_type}'")
@@ -462,7 +494,9 @@ class BaseExecutor(Module):
     def _task_done_callback(self,
                             processing_fut: asyncio.Future,
                             finalizer: Coroutine,
-                            task_future: asyncio.Future
+                            task_future: asyncio.Future,
+                            task: Task,
+                            start_time: float
                             ):
         """
         Handles value returned by task processing future.
@@ -487,14 +521,21 @@ class BaseExecutor(Module):
                 self.logger.fatal(f"Can't handle task result: {exc.__class__.__name__}({exc})")
                 self.harakiri(exc)
 
+            tasks_processed.labels(task_type=task.task_type).inc()
+            tasks_time.labels(task_type=task.task_type).observe(time() - start_time)
+
             # otherwise pass processing future result to task result
             if processing_fut.cancelled():
                 task_future.cancel()
+                tasks_cancelled.labels(task_type=task.task_type).inc()
             elif processing_fut.exception():
                 task_future.set_exception(processing_fut.exception())
+                tasks_failed.labels(task_type=task.task_type).inc()
             else:
                 task_future.set_result(processing_fut.result())
+                tasks_success.labels(task_type=task.task_type).inc()
 
+        tasks_in_progress.labels(task_type=task.task_type).dec()
         handle_fut = self.app_loop.create_task(finalizer)
         handle_fut.add_done_callback(_callback)
 
@@ -523,7 +564,7 @@ class BaseExecutor(Module):
 
         return chain.from_iterable(map(itemgetter(1), handlers))
 
-    async def _handle_task_result(self, task: Task, receiver: AsyncSubscriber, processing_fut: asyncio.Future):
+    async def _handle_task_result(self, task: FrozenTask, receiver: AsyncSubscriber, processing_fut: asyncio.Future):
         """
         Handles task future result.
         :param task:
@@ -567,7 +608,7 @@ class BaseExecutor(Module):
             self.logger.info(f"Task '{task.task_type} (uuid={task.uuid})' processing completed")
         await self._commit_task(task, receiver)
 
-    async def _handle_task_control(self, task: Task, exc: Exception):
+    async def _handle_task_control(self, task: FrozenTask, exc: Exception):
         """
         Takes appropriate action for the task depending on given task control exception.
         This method commits task.
@@ -588,7 +629,7 @@ class BaseExecutor(Module):
             if self.worker.debug:
                 raise ValueError("unknown task control exception")
 
-    async def _commit_task(self, task: Task, receiver: AsyncSubscriber):
+    async def _commit_task(self, task: FrozenTask, receiver: AsyncSubscriber):
         """
         Commits the task on the underlying queue.
         :param task:
