@@ -2,11 +2,12 @@
 import asyncio
 
 import pytest
+import pytest_asyncio
 
 from patchwork.node.executor.unit import InlineUnit
 
 
-@pytest.fixture
+@pytest_asyncio.fixture
 async def unit():
     unit = InlineUnit(parent=None)
     await unit.run()
